import unittest

import numpy as np

from neuron_morphology.morphology import Morphology
from neuron_morphology.constants import (
    AXON, APICAL_DENDRITE, BASAL_DENDRITE, SOMA
)
from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.features.branching import bifurcations as bf
from neuron_morphology.feature_extractor.feature_extractor import FeatureExtractor
from neuron_morphology.feature_extractor.feature_specialization import (
    AxonSpec,
    ApicalDendriteSpec,
    BasalDendriteSpec,
)
from neuron_morphology.feature_extractor.marked_feature import specialize
from neuron_morphology.morphology_builder import MorphologyBuilder


class TestOuterBifurcations(unittest.TestCase):

    def setUp(self):

        self.one_dim_neuron = Morphology([
                {
                    "id": 0,
                    "parent_id": -1,
                    "type": SOMA,
                    "x": 0,
                    "y": 0,
                    "z": 100,
                    "radius": 1
                },
                {
                    "id": 1,
                    "parent_id": 0,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 101,
                    "radius": 1
                },
                {
                    "id": 2,
                    "parent_id": 0,
                    "type": APICAL_DENDRITE,
                    "x": 0,
                    "y": 0,
                    "z": 102,
                    "radius": 1
                },
                { # bifurcates and is within 120
                    "id": 3,
                    "parent_id": 1,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 110,
                    "radius": 1
                },                
                { # This is the farthest node from the root
                    "id": 4,
                    "parent_id": 3,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 140,
                    "radius": 1 
                },                
                { # bifurcates, and is beyond 120
                    "id": 5,
                    "parent_id": 3,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 130,
                    "radius": 1 
                },                
                {
                    "id": 6,
                    "parent_id": 5,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 135,
                    "radius": 1 
                },       
                {
                    "id": 7,
                    "parent_id": 5,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 136,
                    "radius": 1 
                },     
                { # bifurcates and is beyond 120
                    "id": 8,
                    "parent_id": 2,
                    "type": APICAL_DENDRITE,
                    "x": 0,
                    "y": 0,
                    "z": 125,
                    "radius": 1 
                },
                {
                    "id": 9,
                    "parent_id": 8,
                    "type": APICAL_DENDRITE,
                    "x": 0,
                    "y": 0,
                    "z": 126,
                    "radius": 1 
                },
                {
                    "id": 10,
                    "parent_id": 8,
                    "type": APICAL_DENDRITE,
                    "x": 0,
                    "y": 0,
                    "z": 127,
                    "radius": 1 
                },
            ],
            node_id_cb=lambda node: node["id"],
            parent_id_cb=lambda node: node["parent_id"],
        )

        self.one_dim_neuron_data = Data(self.one_dim_neuron)

        self.neurite_features = specialize(
            bf.num_outer_bifurcations,
            {AxonSpec, ApicalDendriteSpec, BasalDendriteSpec}
        )

    def extract(self, feature):
        extractor = FeatureExtractor([feature])
        return (
            extractor.extract(self.one_dim_neuron_data)
            .results
        )

    def test_calculate_outer_bifs(self):
        self.assertEqual(bf.calculate_outer_bifs(
            self.one_dim_neuron,
            self.one_dim_neuron.get_root(),
            None
        ), 2)

    def test_num_outer_bifurcations(self):
        self.assertEqual(
            self.extract(bf.num_outer_bifurcations)["num_outer_bifurcations"],
            2
        )

    def test_apical_num_outer_bifurcations(self):
        self.assertEqual(
            self.extract(self.neurite_features)["apical_dendrite.num_outer_bifurcations"],
            1
        )

    def test_axon_num_outer_bifurcations(self):
        self.assertEqual(
            self.extract(self.neurite_features)["axon.num_outer_bifurcations"],
            1
        )

    def test_basal_num_outer_bifurcations(self):
        # skipped due to no basal dendrite nodes
        self.assertNotIn(
            "basal_dendrite.num_outer_bifurcations",
            self.extract(self.neurite_features),
        )


class TestBifurcationAngles(unittest.TestCase):

    def setUp(self):
        # 180 deg at soma
        # axon - 90 deg local, 60 deg remote
        # basal - 90 deg local, 90 deg remote
        self.morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .axon(0, 0, 1)
                        .axon(0, 1, 2)
                            .axon(0, 1, 1 + np.sqrt(3)).up(2)
                        .axon(0, -1, 2)
                            .axon(0, -1, 1 + np.sqrt(3)).up(3)
                    .basal_dendrite(0, 0, -1)
                        .basal_dendrite(0, 1, -2).up()
                        .basal_dendrite(0, -1, -2)
                .build()
        )
        nodes = self.morphology.nodes()
        for node in nodes:
            print(node)
        self.data = Data(self.morphology)
        self.local_specialized = specialize(
            bf.mean_bifurcation_angle_local,
            {AxonSpec, ApicalDendriteSpec, BasalDendriteSpec}
        )
        self.remote_specialized = specialize(
            bf.mean_bifurcation_angle_remote,
            {AxonSpec, ApicalDendriteSpec, BasalDendriteSpec}
        )

    def extract(self, feature):
        extractor = FeatureExtractor([feature])
        return (
            extractor.extract(self.data)
            .results
        )

    def test_mean_bifurcation_angle_local(self):
        self.assertAlmostEqual(
            self.extract(bf.mean_bifurcation_angle_local)["mean_bifurcation_angle_local"],
            (np.pi / 2 + np.pi / 2 + np.pi) / 3
        )

    def test_mean_bifurcation_angle_remote(self):
        self.assertAlmostEqual(
            self.extract(bf.mean_bifurcation_angle_remote)["mean_bifurcation_angle_remote"],
            (np.pi / 2 + np.pi / 3 + np.pi) / 3
        )

    def test_axon_mean_bifurcation_angle_local(self):
        self.assertAlmostEqual(
            self.extract(self.local_specialized)["axon.mean_bifurcation_angle_local"],
            np.pi / 2
        )

    def test_axon_mean_bifurcation_angle_remote(self):
        self.assertAlmostEqual(
            self.extract(self.remote_specialized)["axon.mean_bifurcation_angle_remote"],
            np.pi / 3
        )



import os
import psycopg2
import numpy as np


def convert_coords_str(coords_str: str, resolution=None):
    """Convert a comma seperated string of coordinate pairs"""
    vals = coords_str.split(',')

    if resolution:
        resolution = float(resolution)
    else:
        resolution = 1.0

    x = np.asarray(vals[0::2], dtype=float) * resolution
    y = np.asarray(vals[1::2], dtype=float) * resolution
    coords = {'x': x, 'y': y}

    return coords


def get_data(query):
    host = os.getenv("DBHOST")
    dbname = os.getenv("DBNAME")
    user = os.getenv("DBREADER")
    password = os.getenv("DBPASSWORD")
    conn_str = f'host={host} dbname={dbname} user={user} password={password}'

    data = {}
    with psycopg2.connect(conn_str) as conn:
        conn.set_session(readonly=True)
        with conn.cursor() as curs:
            curs.execute(query)
            for row in curs:
                (specimen_id, apical_tag, path_type,
                 path_str, resolution, swc_dir, swc_name) = row
                if specimen_id not in data:
                    data[specimen_id] = {'apical_tag': apical_tag,
                                         'resolution': resolution,
                                         'swc_file': os.path.join(swc_dir, swc_name),
                                         'soma_path_str': None,
                                         'soma_center': None,
                                         'soma_center_pix': None,
                                         'pia_path_str': None,
                                         'pia_coords': None,
                                         'wm_path_str': None,
                                         'wm_coords': None,
                                         }
                if path_type == 'Soma':
                    data[specimen_id]['soma_path_str'] = path_str
                    soma_coords = convert_coords_str(path_str,  resolution=data[specimen_id]['resolution'])
                    data[specimen_id]['soma_center'] = np.asarray(
                        (soma_coords['x'].mean(), soma_coords['y'].mean()))
                    soma_pix_coords = convert_coords_str(path_str)
                    data[specimen_id]['soma_center_pix'] = np.asarray(
                        (soma_pix_coords['x'].mean(), soma_pix_coords['y'].mean()))
                elif path_type == 'Pia':
                    data[specimen_id]['pia_path_str'] = path_str
                    data[specimen_id]['pia_coords'] = convert_coords_str(path_str, resolution=data[specimen_id]['resolution'])
                elif path_type == 'White Matter':
                    data[specimen_id]['wm_path_str'] = path_str
                    data[specimen_id]['wm_coords'] = convert_coords_str(path_str, resolution=data[specimen_id]['resolution'])

    invalid_ids = []
    for sid, sdata in data.items():
        if (sdata['soma_path_str'] is None or
                sdata['pia_path_str'] is None or
                sdata['wm_path_str'] is None):

            invalid_ids.append(sid)

    invalid_data = {sid: data.pop(sid) for sid in invalid_ids}
    return data, invalid_data


def get_all_intact_apical():
    query = """
        SELECT DISTINCT sts.specimen_id as specimen_id, specimen_tags.name, layert.name as path_type, poly.path, sc.resolution, wkf.storage_directory, wkf.filename
        FROM specimen_tags_specimens AS sts
        JOIN specimen_tags ON specimen_tags.id=sts.specimen_tag_id
        AND specimen_tags.name IN ('apical - intact')
        JOIN biospecimen_polygons AS bsp ON bsp.biospecimen_id=sts.specimen_id
        JOIN avg_graphic_objects poly ON poly.id=bsp.polygon_id
        JOIN avg_graphic_objects layer ON layer.id=poly.parent_id
        JOIN avg_group_labels layert ON layert.id=layer.group_label_id
        AND layert.prevent_missing_polygon_structure=false
        JOIN sub_images AS si ON si.id=layer.sub_image_id
        AND si.failed=false
        JOIN images AS im ON im.id=si.image_id
        JOIN slides AS s ON s.id=im.slide_id
        JOIN scans AS sc ON sc.slide_id=s.id
        AND sc.superseded=false
        JOIN treatments t ON t.id = im.treatment_id AND t.id = 300080909 --Why?
        JOIN neuron_reconstructions AS nr ON nr.specimen_id=sts.specimen_id
        AND nr.superseded=false
        JOIN well_known_files AS wkf ON wkf.attachable_id=nr.id
        AND wkf.attachable_type='NeuronReconstruction'
        JOIN well_known_file_types AS wkft on wkft.id=well_known_file_type_id
        AND wkft.name='3DNeuronReconstruction'
        ORDER BY sts.specimen_id
    """
    return get_data(query)


def get_benchmark_apical():
    # From Nathan Gouwens
    # neurons with m-type SPINY 3,4,5,7,8,9,10,11,12,13,14,16,17
    specimen_ids = [
     313862022, 314642645, 314804042, 314822529, 314831019, 314900022, 318733871,
     318808427, 320207387, 320668879, 321906005, 323865917, 324025297, 324025371,
     324521027, 325941643, 329550137, 348592897, 349621141, 354190013, 354833767,
     370351753, 382982932, 386049446, 386970660, 395830185, 397351623, 397353539,
     422738880, 466632464, 467703703, 468120757, 468193142, 469704261, 469793303,
     469798159, 471129934, 471141261, 471143169, 471758398, 471767045, 471786879,
     471789504, 471800189, 473020156, 473601979, 473611755, 475459689, 475744706,
     476048909, 476049169, 476054887, 476086391, 476126528, 476131588, 476135066,
     476216637, 476218657, 476263004, 476451456, 476455864, 476457450, 476562817,
     476616076, 476823462, 477127614, 477135941, 478497530, 478586295, 478888052,
     478892782, 479010903, 479013100, 479179020, 479225052, 479225080, 479704527,
     479721491, 479728896, 479770916, 480003970, 480087928, 480114344, 480116737,
     480122859, 480124551, 480169178, 480351780, 480353286, 483020137, 483061182,
     483068687, 483092310, 484564503, 484679812, 485161419, 485468180, 485574721,
     485574832, 485835016, 485835055, 485836906, 485837504, 485838981, 485880739,
     485909730, 485912047, 485931158, 486025194, 486052980, 486110216, 486111903,
     486132712, 486146717, 486239338, 486502127, 486893033, 487601493, 487664663,
     488117124, 488420599, 488679042, 488680917, 488683425, 488695444, 488697163,
     488698341, 490205998, 490259231, 490485142, 490916882, 497611660, 501799874,
     501847931, 501956013, 502267531, 502269786, 502366405, 502614426, 502978383,
     502999078, 503286448, 503814817, 507101551, 507918877, 509003464, 510658021,
     513800575, 515249852, 515524026, 515986607, 517077548, 521938313, 524850271,
     527826878, 555241875, 556369531, 557037024, 557203722, 557252022, 557864274,
     557874460, 557998843, 560678143, 563350008, 566506157, 568568071, 569653118,
     588402092, 588419063, 588712191, 589128331, 589427435, 589442285, 589760138,
     598628992, 599334696, 602822298, 607124114, 656411052, 656421538, 656425408,
     656658253, 656707861, 676322245, 698217104, 698226504, 698230720, 698235442,
     698237629
     ]

    ids_str = ', '.join([str(sid) for sid in specimen_ids])

    query = f"""
        SELECT DISTINCT sts.specimen_id as specimen_id, specimen_tags.name, layert.name as path_type, poly.path, sc.resolution, wkf.storage_directory, wkf.filename
        FROM specimen_tags_specimens AS sts
        JOIN specimen_tags ON specimen_tags.id=sts.specimen_tag_id
        AND sts.specimen_id IN ({ids_str})
        AND specimen_tags.name IN ('apical - intact')
        JOIN biospecimen_polygons AS bsp ON bsp.biospecimen_id=sts.specimen_id
        JOIN avg_graphic_objects poly ON poly.id=bsp.polygon_id
        JOIN avg_graphic_objects layer ON layer.id=poly.parent_id
        JOIN avg_group_labels layert ON layert.id=layer.group_label_id
        AND layert.prevent_missing_polygon_structure=false
        JOIN sub_images AS si ON si.id=layer.sub_image_id
        AND si.failed=false
        JOIN images AS im ON im.id=si.image_id
        JOIN slides AS s ON s.id=im.slide_id
        JOIN scans AS sc ON sc.slide_id=s.id
        AND sc.superseded=false
        JOIN treatments t ON t.id = im.treatment_id AND t.id = 300080909 --Why?
        JOIN neuron_reconstructions AS nr ON nr.specimen_id=sts.specimen_id
        AND nr.superseded=false
        JOIN well_known_files AS wkf ON wkf.attachable_id=nr.id
        AND wkf.attachable_type='NeuronReconstruction'
        JOIN well_known_file_types AS wkft on wkft.id=well_known_file_type_id
        AND wkft.name='3DNeuronReconstruction'
        ORDER BY sts.specimen_id
        """
    return get_data(query)



from tests.objects import test_marker, test_node, test_tree
from neuron_morphology.validation import marker_validation as mv
from tests.validation.validation_test_case import ValidationTestCase
from neuron_morphology.constants import *
import unittest
from mock import patch


class TestMarkerValidationFunctions(ValidationTestCase):
    """ Tests the functions in marker_validation.py """

    def test_validate_expected_name_valid(self):
        for marker_name in [CUT_DENDRITE, NO_RECONSTRUCTION, TYPE_30]:
            errors = mv.validate_expected_name([test_marker(name=marker_name)])
            self.assertEqual(len(errors), 0)

    def test_validate_expected_name_invalid(self):
        errors = mv.validate_expected_name([test_marker(name=5)])
        self.assertMarkerErrors(errors, "Marker name needs to be one of these values:",
                                [[{'x': 0, 'y': 0, 'z': 0, 'name': 5}]])

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_coordinate_corresponding_to_dendrite_tips_cut_dendrite_valid(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, x=0, y=0, z=0, parent_node_id=1)]
            mv.validate([test_marker(x=0, y=0, z=0, name=CUT_DENDRITE)], test_tree(nodes))

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_coordinate_corresponding_to_dendrite_tips_cut_dendrite_invalid(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, x=0, y=0, z=0, parent_node_id=1)]
            errors = mv.validate([test_marker(x=1, y=0, z=0, name=CUT_DENDRITE),
                                  test_marker(x=1, y=0, z=0, name=TYPE_30)], test_tree(nodes))

            self.assertMarkerErrors(errors, "Coordinates for each dendrite (type 10) needs to correspond to "
                                            "a tip of a dendrite type (type 3 or 4) in the related morphology",
                                    [[{'x': 1, 'y': 0, 'z': 0, 'name': CUT_DENDRITE}]])

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_coordinate_corresponding_to_dendrite_tips_multiple_cut_dendrite_invalid(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, x=0, y=0, z=0, parent_node_id=1),
                     test_node(id=2, type=dendrite_type, x=0, y=0, z=0, parent_node_id=1)]
            errors = mv.validate([test_marker(x=1, y=0, z=0, name=CUT_DENDRITE),
                                  test_marker(x=1, y=0, z=0, name=TYPE_30)], test_tree(nodes))

            self.assertMarkerErrors(errors, "Coordinates for each dendrite (type 10) needs to correspond to "
                                            "a tip of a dendrite type (type 3 or 4) in the related morphology",
                                    [[{'x': 1, 'y': 0, 'z': 0, 'name': CUT_DENDRITE}]])

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_coordinate_corresponding_to_axon_tips_no_reconstruction_invalid(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                 test_node(id=2, type=AXON, x=0, y=0, z=0, parent_node_id=1)]
        errors = mv.validate([test_marker(x=1, y=0, z=0, name=NO_RECONSTRUCTION),
                              test_marker(x=1, y=0, z=0, name=TYPE_30)], test_tree(nodes))

        self.assertMarkerErrors(errors, "Coordinates for each axon (type 20) needs to correspond to "
                                        "a tip of an axon type (type 2) in the related morphology",
                                [[{'x': 1, 'y': 0, 'z': 0, 'name': NO_RECONSTRUCTION}]])

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_coordinate_corresponding_to_dendrite_tips_type_20_valid(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, x=0, y=0, z=0, parent_node_id=1)]
            mv.validate([test_marker(x=1, y=0, z=0, name=NO_RECONSTRUCTION)], test_tree(nodes))

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_coordinate_corresponding_to_dendrite_tips_type_30_valid(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, x=0, y=0, z=0, parent_node_id=1)]
            mv.validate([test_marker(x=1, y=0, z=0, name=TYPE_30)], test_tree(nodes))

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_coordinate_corresponding_to_axon_tips_type_20_valid(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                 test_node(id=2, type=AXON, x=0, y=0, z=0, parent_node_id=1)]
        mv.validate([test_marker(x=0, y=0, z=0, name=AXON)], test_tree(nodes))

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_coordinate_corresponding_to_dendrite_tips_type_20_valid(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, x=0, y=0, z=0, parent_node_id=1)]
            mv.validate([test_marker(x=1, y=0, z=0, name=AXON)], test_tree(nodes))

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_number_of_type_30_one(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1)]
        mv.validate([test_marker(name=TYPE_30)], test_tree(nodes))

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_number_of_type_30_less_than_one(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1)]
        errors = mv.validate([test_marker(name=NO_RECONSTRUCTION)], test_tree(nodes))
        self.assertMarkerErrors(errors, "Total number of type 30s is 0", [[{}]])

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_number_of_type_30_more_than_one(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1)]
        errors = mv.validate([test_marker(name=TYPE_30), test_marker(name=TYPE_30)], test_tree(nodes))
        self.assertMarkerErrors(errors, "Total number of type 30s is 2", [[{'x': 0, 'y': 0, 'z': 0, 'name': TYPE_30}],
                                                                          [{'x': 0, 'y': 0, 'z': 0, 'name': TYPE_30}]])

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_number_of_no_reconstruction_one(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1)]
        mv.validate([test_marker(name=NO_RECONSTRUCTION)], test_tree(nodes))

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_number_of_no_reconstruction_zero(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1)]
        mv.validate([test_marker(name=CUT_DENDRITE)], test_tree(nodes))

    @patch("neuron_morphology.validation.marker_validators", [mv])
    def test_number_of_no_reconstruction_more_than_one(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1)]
        errors = mv.validate([test_marker(name=NO_RECONSTRUCTION), test_marker(name=NO_RECONSTRUCTION)],
                             test_tree(nodes))
        self.assertMarkerErrors(errors, "Total number of type 20s is more than one: 2",
                                [[{'x': 0, 'y': 0, 'z': 0, 'name': NO_RECONSTRUCTION}],
                                 [{'x': 0, 'y': 0, 'z': 0, 'name': NO_RECONSTRUCTION}]])


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestMarkerValidationFunctions)
    unittest.TextTestRunner(verbosity=5).run(suite)



import unittest
from unittest import TestCase
from unittest.mock import MagicMock, patch
import tempfile
import shutil
import os
import sys

import pytest
import numpy as np
import imageio
import glymur
import pandas as pd
import matplotlib.pyplot as plt

try:
    import rasterio
    import neuron_morphology.snap_polygons.image_outputter as imgo
    from neuron_morphology.snap_polygons.geometries import Geometries
except ImportError:
    pass

@unittest.skipIf('rasterio' not in sys.modules,
                 'install rasterio to use snap_polygons.geometries')
class TestUtilities(TestCase):

    def test_make_path_patch_open(self):
        vertices = [(0, 0), (1, 0), (1, 1)]
        alpha = 0.1

        obtained = imgo.make_pathpatch(vertices, alpha=alpha)

        obt_path = obtained.get_path()
        obt_verts = obt_path.vertices
        obt_alpha = obtained.get_alpha()

        assert np.allclose(obt_verts, vertices)
        self.assertAlmostEqual(obt_alpha, alpha)

    def test_fname_suffix(self):
        path = "/a/b/c.foo"
        suffix = "baz"

        obt = imgo.fname_suffix(path, suffix)
        self.assertEqual(obt, "/a/b/c_baz.foo")

@unittest.skipIf('rasterio' not in sys.modules,
                 'install rasterio to use snap_polygons.geometries')
class TestImageIo(TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def test_read_png(self):
        path = os.path.join(self.tmpdir, "test.png")
        image = np.arange(16, dtype=np.uint8).reshape(4, 4)
        imageio.imsave(path, image)

        obt = imgo.read_image(path)
        assert np.allclose(obt, image)

    def test_read_png_dec(self):
        path = os.path.join(self.tmpdir, "test_dec.png")
        image = np.arange(16, dtype=np.uint8).reshape(4, 4)
        imageio.imsave(path, image)

        obt = imgo.read_image(path, 2)
        assert np.allclose(
            obt, 
            np.array([
                [0, 2],
                [8, 10]
            ])
        )

    @pytest.mark.skipif( # we are running via pytest
        glymur.version.openjpeg_version < "2.3.0", 
        reason="you must have at openjpeg of at least 2.3.0 to write images"
    )
    def test_read_jp2(self):
        path = os.path.join(self.tmpdir, "test.jp2")
        image = np.random.randint(0, 255, (640, 480), dtype=np.uint8)
        glymur.Jp2k(path, data=image)

        obt = imgo.read_image(path)
        assert np.allclose(obt, image)


    def test_write_figure(self):
        fig, ax = plt.subplots()
        ax.plot(np.arange(2))
        fig2, ax2 = plt.subplots()

        path = os.path.join(self.tmpdir, "test.png")
        path2 = os.path.join(self.tmpdir, "test2.png")

        imgo.write_figure(fig2, path2)
        imgo.write_figure(fig, path)

        self.assertEqual(plt.gcf().number, fig2.number) # current fig is reset
        assert os.stat(path).st_size > os.stat(path2).st_size # correct fig is written

@unittest.skipIf('rasterio' not in sys.modules,
                 'install rasterio to use snap_polygons.geometries')
class TestImageOutputter(TestCase):
    # ImageOutputter has a lot of "does it look right" plotting logic. We 
    # won't be testing that, but we will make sure the overall behavior is OK


    def setUp(self):
        self.native_layer1_path =  [(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)]
        self.native = Geometries()
        self.native.register_polygon("layer1", self.native_layer1_path)

        self.result_layer1_path = [(0, 0), (0, 10), (10, 10), (10, 0), (0, 0)]
        self.result_layer1_pia = [(10, 10), (10, 0)]
        self.result = Geometries()
        self.result.register_polygon("layer1", self.result_layer1_path)
        self.result.register_surface("layer1_pia", self.result_layer1_pia)

        self.image_specs = [
            {
                "input_path": "fizz", 
                "output_path": "buzz", 
                "downsample": 8, 
                "overlay_types": ["before", "after"]
            }
        ]

        self.outputter = imgo.ImageOutputter(
            self.native, self.result, self.image_specs)

    def test_draw_before(self):
        fig = self.outputter.draw_before(np.eye(3))
        ax = fig.get_axes()[0]
        vertices = ax.patches[0].get_path().vertices
        assert np.allclose(vertices, self.native_layer1_path)

    def test_draw_after(self):
        fig = self.outputter.draw_after(np.eye(30))
        ax = fig.get_axes()[0]

        poly_vertices = ax.patches[0].get_path().vertices
        assert np.allclose(poly_vertices, self.result_layer1_path)

        surf_vertices = ax.patches[1].get_path().vertices
        assert np.allclose(surf_vertices, self.result_layer1_pia)

    def test_write_images(self):
        with patch(
            "neuron_morphology.snap_polygons.image_outputter.write_figure", 
            new_callable=MagicMock
        ) as write:
            with patch(
                "neuron_morphology.snap_polygons.image_outputter.read_image",
                new=lambda *a, **k: np.eye(30)
            ) as _read:
                obtained = self.outputter.write_images()
                write.assert_called()

        pd.testing.assert_frame_equal(
            pd.DataFrame(obtained), 
            pd.DataFrame({
                "input_path": ["fizz", "fizz"], 
                "downsample": [8, 8],
                "output_path": ["buzz_before", "buzz_after"],
                "overlay_type": ["before", "after"]
            }),
            check_like=True,
            check_dtype=False
        )


from operator import add
from neuron_morphology.constants import *

# TODO this is a port of an older statistics module. We should roll it into the 
# feature extractor.


def count_number_of_independent_axons(morphology):

    """ This functions counts the number of independent axons (parent is -1) """

    count = 0
    axon_nodes = morphology.get_node_by_types([AXON])

    for node in axon_nodes:
        if morphology.parent_of(node) is None:
            count += 1

    return count


def morphology_statistics(morphology):
    return {
        "Number of Independent Axons": count_number_of_independent_axons(morphology)
    }


import marshmallow as mm

from argschema import ArgSchema
from argschema.schemas import DefaultSchema
from argschema.fields import Nested, Float, List, InputFile, OutputFile


def validate_input_affine(data):
    has_affine_dict = bool(data.get("affine_dict"))
    has_affine_list = bool(data.get("affine_list"))
    if has_affine_list and has_affine_dict:
        err_msg = "Provide either affine_list or affine_dict, not both"
        raise mm.ValidationError(err_msg)
    elif not has_affine_list and not has_affine_dict:
        err_msg = "Provide either affine_list or affine_dict"
        raise mm.ValidationError(err_msg)


class AffineDictSchema(DefaultSchema):
    """
    affine_dict: keys and values corresponding to the following
                [[tvr_00 tvr_01 tvr_02 tvr_09]
                 [tvr_03 tvr_04 tvr_05 tvr_10]
                 [tvr_06 tvr_07 tvr_08 tvr_11]
                 [0      0      0      1]]
    """

    tvr_00 = Float(required=True)
    tvr_01 = Float(required=True)
    tvr_02 = Float(required=True)
    tvr_03 = Float(required=True)
    tvr_04 = Float(required=True)
    tvr_05 = Float(required=True)
    tvr_06 = Float(required=True)
    tvr_07 = Float(required=True)
    tvr_08 = Float(required=True)
    tvr_09 = Float(required=True)
    tvr_10 = Float(required=True)
    tvr_11 = Float(required=True)


class ApplyAffineSchema(ArgSchema):
    """Arg Schema for apply_affine_transform module"""

    affine_dict = Nested(
        AffineDictSchema,
        required=False,
        description="Dictionary defining an affine transform",
    )
    affine_list = List(
        Float,
        required=False,
        cli_as_single_argument=True,
        description="List defining an affine transform",
    )
    input_swc = InputFile(required=True, description="swc file to be transformed")

    output_swc = OutputFile(required=True, description="Output swc filepath")

    @mm.validates_schema
    def validate_schema_input(self, data, **kwargs):
        validate_input_affine(data)


class OutputParameters(DefaultSchema):
    inputs = Nested(
        ApplyAffineSchema,
        description="The parameters argued to this executable",
        required=True,
    )
    transformed_swc = OutputFile(
        required=True, description="location of the transformed swc"
    )
